@@ -1,9 +1,10 @@
 /*
- * Copyright (C) 2010 Red Hat, Inc.
+ * Copyright (c) 2017 Red Hat, Inc.
  *
  * All rights reserved.
  *
  * Author: Angus Salkeld <asalkeld@redhat.com>
+ *         Jan Pokorny <jpokorny@redhat.com>
  *
  * libqb is free software: you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -38,10 +39,6 @@ extern "C" {
 #include <qb/qbutil.h>
 #include <qb/qbconfig.h>
 
-#ifdef S_SPLINT_S
-#undef QB_HAVE_ATTRIBUTE_SECTION
-#endif /* S_SPLINT_S */
-
 /**
  * @file qblog.h
  * The logging API provides four main parts (basics, filtering, threading & blackbox).
@@ -66,7 +63,12 @@ extern "C" {
  *
  * @par Configuring log targets.
  * A log target can be syslog, stderr, the blackbox, stdout, or a text file.
- * By default only syslog is enabled.
+ * By default, only syslog is enabled.  While this is usual for daemons,
+ * it is rarely appropriate for ordinary programs, which should
+ * disable it when other targets (see below) are to be used:
+ * @code
+ *	qb_log_ctl(B_LOG_SYSLOG, QB_LOG_CONF_ENABLED, QB_FALSE);
+ * @endcode
  *
  * To enable a target do the following:
  * @code
@@ -74,16 +76,16 @@ extern "C" {
  * @endcode
  *
  * syslog, stderr, the blackbox, and stdout are static (they don't need
- * to be created, just enabled or disabled). However you can open multiple
- * logfiles (QB_LOG_TARGET_MAX - QB_LOG_TARGET_STATIC_MAX).
- * To do this, use the following code:
+ * to be created, just enabled or disabled).  However, you can open multiple
+ * logfiles (falling within inclusive range @c QB_LOG_TARGET_DYNAMIC_START
+ * up to @c QB_LOG_TARGET_DYNAMIC_END).  To do this, use the following code:
  * @code
  *	mytarget = qb_log_file_open("/var/log/mylogfile");
  *	qb_log_ctl(mytarget, QB_LOG_CONF_ENABLED, QB_TRUE);
  * @endcode
  *
- * Once your targets are enabled/opened you can configure them as follows:
- * Configure the size of blackbox
+ * Once your targets are enabled/opened, you can configure them as follows:
+ * Configure the size of blackbox:
  * @code
  *	qb_log_ctl(QB_LOG_BLACKBOX, QB_LOG_CONF_SIZE, 1024*10);
  * @endcode
@@ -93,13 +95,21 @@ extern "C" {
  *	qb_log_ctl(mytarget, QB_LOG_CONF_THREADED, QB_TRUE);
  * @endcode
  *
- * To workaround your syslog daemon filtering all messages > LOG_INFO
+ * Sometimes, syslog daemons are (pre)configured to filter messages not
+ * exceeding a particular priority.  When this happens to be the logging
+ * target, the designated priority of the message is passed along unchanged,
+ * possibly resulting in message loss.  For messages up to @c LOG_DEBUG
+ * importance, this can be worked around by proportionally bumping the
+ * priorities to be passed to syslog (here, the step is such that
+ * @c LOG_DEBUG gets promoted to @c LOG_INFO):
  * @code
  *	qb_log_ctl(QB_LOG_SYSLOG, QB_LOG_CONF_PRIORITY_BUMP,
- *		   LOG_INFO - LOG_DEBUG);
+ *	           LOG_INFO - LOG_DEBUG);
  * @endcode
  *
- * To ensure all logs to file targets are fsync'ed (default QB_FALSE)
+ * To ensure all logs to file targets are fsync'ed (new messages expressly
+ * transferred to the storage device as they keep coming, otherwise defaults
+ * to @c QB_FALSE):
  * @code
  *	qb_log_ctl(mytarget, QB_LOG_CONF_FILE_SYNC, QB_TRUE);
  * @endcode
@@ -232,14 +242,13 @@ extern "C" {
 #define LOG_TRACE    (LOG_DEBUG + 1)
 
 #define QB_LOG_MAX_LEN 512
+#define QB_LOG_ABSOLUTE_MAX_LEN 4096
 #define QB_LOG_STRERROR_MAX_LEN 128
 
 typedef const char *(*qb_log_tags_stringify_fn)(uint32_t tags);
 
 /**
- * An instance of this structure is created in a special
- * ELF section at every dynamic debug callsite.  At runtime,
- * the special section is treated as an array of these.
+ * An instance of this structure is created for each log message
  */
 struct qb_log_callsite {
 	const char *function;
@@ -253,30 +262,7 @@ struct qb_log_callsite {
 
 typedef void (*qb_log_filter_fn)(struct qb_log_callsite * cs);
 
-/* will be assigned by linker magic (assuming linker supports that):
- * https://sourceware.org/binutils/docs/ld/Orphan-Sections.html
- */
-#ifdef QB_HAVE_ATTRIBUTE_SECTION
-
-#define QB_ATTR_SECTION			__verbose  /* conforms to C ident. */
-#define QB_ATTR_SECTION_STR		QB_PP_STRINGIFY(QB_ATTR_SECTION)
-#define QB_ATTR_SECTION_START		QB_PP_JOIN(__start_, QB_ATTR_SECTION)
-#define QB_ATTR_SECTION_STOP		QB_PP_JOIN(__stop_, QB_ATTR_SECTION)
-#define QB_ATTR_SECTION_START_STR	QB_PP_STRINGIFY(QB_ATTR_SECTION_START)
-#define QB_ATTR_SECTION_STOP_STR	QB_PP_STRINGIFY(QB_ATTR_SECTION_STOP)
-
-extern struct qb_log_callsite QB_ATTR_SECTION_START[];
-extern struct qb_log_callsite QB_ATTR_SECTION_STOP[];
-
-/* mere linker sanity check, possible future extension for internal purposes */
-#define QB_LOG_INIT_DATA(name)							\
-    void name(void);								\
-    void name(void) 								\
-    { if (QB_ATTR_SECTION_START == QB_ATTR_SECTION_STOP) assert(0); }	\
-    void __attribute__ ((constructor)) name(void);
-#else
 #define QB_LOG_INIT_DATA(name)
-#endif
 
 /**
  * Internal function: use qb_log() or qb_logt()
@@ -285,7 +271,7 @@ void qb_log_real_(struct qb_log_callsite *cs, ...);
 void qb_log_real_va_(struct qb_log_callsite *cs, va_list ap);
 
 #define QB_LOG_TAG_LIBQB_MSG_BIT 31
-#define QB_LOG_TAG_LIBQB_MSG (1 << QB_LOG_TAG_LIBQB_MSG_BIT)
+#define QB_LOG_TAG_LIBQB_MSG (1U << QB_LOG_TAG_LIBQB_MSG_BIT)
 
 /**
  * This function is to import logs from other code (like libraries)
@@ -354,21 +340,12 @@ void qb_log_from_external_source_va(const char *function,
  * @param fmt usual printf style format specifiers
  * @param args usual printf style args
  */
-#ifdef QB_HAVE_ATTRIBUTE_SECTION
-#define qb_logt(priority, tags, fmt, args...) do {			\
-	static struct qb_log_callsite descriptor			\
-	__attribute__((section(QB_ATTR_SECTION_STR), aligned(8))) =	\
-	{ __func__, __FILE__, fmt, priority, __LINE__, 0, tags };	\
-	qb_log_real_(&descriptor, ##args);				\
-    } while(0)
-#else
 #define qb_logt(priority, tags, fmt, args...) do {	\
 	struct qb_log_callsite* descriptor_pt =		\
 	qb_log_callsite_get(__func__, __FILE__, fmt,	\
 			    priority, __LINE__, tags);	\
 	qb_log_real_(descriptor_pt, ##args);		\
     } while(0)
-#endif /* QB_HAVE_ATTRIBUTE_SECTION */
 
 
 /**
@@ -417,8 +394,8 @@ void qb_log_from_external_source_va(const char *function,
  * as non-inclusive higher bounds of the respective categories
  * (static and all the log targets) and also denote the number
  * of (reserved) items in the category.  Both are possibly subject
- * of change, hence it is only adequate to always refer to them
- * via these defined values.
+ * to change, so you should always refer to them using
+ * these defined values.
  * Similarly, there are QB_LOG_TARGET_{STATIC_,DYNAMIC_,}START
  * and QB_LOG_TARGET_{STATIC_,DYNAMIC_,}END values, but these
  * are inclusive lower and higher bounds, respectively.
@@ -460,6 +437,9 @@ enum qb_log_conf {
 	QB_LOG_CONF_FILE_SYNC,
 	QB_LOG_CONF_EXTENDED,
 	QB_LOG_CONF_IDENT,
+	QB_LOG_CONF_MAX_LINE_LEN,
+	QB_LOG_CONF_ELLIPSIS,
+	QB_LOG_CONF_USE_JOURNAL,
 };
 
 enum qb_log_filter_type {
@@ -482,11 +462,11 @@ enum qb_log_filter_conf {
 
 typedef void (*qb_log_logger_fn)(int32_t t,
 				 struct qb_log_callsite *cs,
-				 time_t timestamp,
+				 struct timespec *timestamp,
 				 const char *msg);
 typedef void (*qb_log_vlogger_fn)(int32_t t,
 				 struct qb_log_callsite *cs,
-				 time_t timestamp,
+				 struct timespec *timestamp,
 				 va_list ap);
 
 typedef void (*qb_log_close_fn)(int32_t t);
@@ -535,7 +515,7 @@ void qb_log_callsites_dump(void);
  *
  * @param target QB_LOG_SYSLOG, QB_LOG_STDERR or result from qb_log_file_open()
  * @param conf_type configuration directive ("what to configure") that accepts
- *        <tt>int32_t</tt> argument determining the new value unless ignored
+ *        @c int32_t argument determining the new value unless ignored
  *        for particular directive altogether
  *        (incompatible directives: QB_LOG_CONF_IDENT)
  * @param arg the new value for a state-changing configuration directive,
@@ -558,7 +538,7 @@ typedef union {
  *
  * @param target QB_LOG_SYSLOG, QB_LOG_STDERR or result from qb_log_file_open()
  * @param conf_type configuration directive ("what to configure") that accepts
- *        either <tt>int32_t</tt> or a null-terminated string argument
+ *        either @c int32_t or a null-terminated string argument
  *        determining the new value unless ignored for particular directive
  *        (compatible directives: those valid for qb_log_ctl
  *                                + QB_LOG_CONF_IDENT)
@@ -569,9 +549,9 @@ typedef union {
  *        that original function directly as it allows for more type safety)
  * @see qb_log_ctl
  *
- * @note You can use <tt>QB_LOG_CTL2_I32</tt> and <tt>QB_LOG_CTL2_S</tt>
- *       macros for a convenient on-the-fly construction of the object
- *       to be passed as an <tt>arg</tt> argument.
+ * @note You can use @ref QB_LOG_CTL2_I32 and @ref QB_LOG_CTL2_S macros
+ *       for a convenient on-the-fly construction of the object
+ *       to be passed as an @p arg argument.
  */
 int32_t qb_log_ctl2(int32_t target, enum qb_log_conf conf_type,
 		    qb_log_ctl2_arg_t arg);
@@ -626,6 +606,15 @@ int32_t qb_log_filter_fn_set(qb_log_filter_fn fn);
  */
 void qb_log_tags_stringify_fn_set(qb_log_tags_stringify_fn fn);
 
+
+/**
+ *This is a Feature Test macro so that calling applications know that
+ * millisecond timestamps are implemented. Because %T a string in
+ * function call with an indirect effect, there is no easy test for it
+ * beyond the library version (which is a very blunt instrument)
+ */
+#define QB_FEATURE_LOG_HIRES_TIMESTAMPS 1
+
 /**
  * Set the format specifiers.
  *
@@ -634,13 +623,21 @@ void qb_log_tags_stringify_fn_set(qb_log_tags_stringify_fn fn);
  * %l FILELINE
  * %p PRIORITY
  * %t TIMESTAMP
+ * %T TIMESTAMP with milliseconds
  * %b BUFFER
  * %g TAGS
  * %N name (passed into qb_log_init)
  * %P PID
  * %H hostname
  *
- * any number between % and character specify field length to pad or chop
+ * Any number between % and character specify field length to pad or chop.
+ *
+ * @note Some of the fields are immediately evaluated and remembered
+ *       for performance reasons, so whenlog messages carry PIDs (not the default)
+ *       this function needs to be reinvoked following @c fork
+ *       (@c clone) in the respective children.  When already linking
+ *       with @c libpthread, @c pthread_atfork callback registration
+ *       could be useful.
  */
 void qb_log_format_set(int32_t t, const char* format);
 
@@ -655,10 +652,20 @@ void qb_log_format_set(int32_t t, const char* format);
 int32_t qb_log_file_open(const char *filename);
 
 /**
- * Close a log file and release is resources.
+ * Close a log file and release its resources.
  */
 void qb_log_file_close(int32_t t);
 
+/**
+ * Open a new log file for an existing target
+ * @param t target
+ * @param filename may be NULL to use existing file name
+ *
+ * @retval -errno on error
+ *
+ */
+int32_t qb_log_file_reopen(int32_t t, const char *filename);
+
 /**
  * When using threaded logging set the pthread policy and priority.
  *
@@ -680,7 +687,7 @@ ssize_t qb_log_blackbox_write_to_file(const char *filename);
 /**
  * Read the blackbox for file and print it out.
  */
-void qb_log_blackbox_print_from_file(const char* filename);
+int qb_log_blackbox_print_from_file(const char* filename);
 
 /**
  * Open a custom log target.
@@ -696,7 +703,7 @@ int32_t qb_log_custom_open(qb_log_logger_fn log_fn,
 			   void *user_data);
 
 /**
- * Close a custom log target and release is resources.
+ * Close a custom log target and release its resources.
  */
 void qb_log_custom_close(int32_t t);
 
@@ -707,19 +714,19 @@ void qb_log_custom_close(int32_t t);
 void *qb_log_target_user_data_get(int32_t t);
 
 /**
- * Associate user data with this log target
+ * Associate user data with this log target.
  * @note only use this with custom targets
  */
 int32_t qb_log_target_user_data_set(int32_t t, void *user_data);
 
 /**
- * format the callsite and timestamp info according to the format
+ * Format the callsite and timestamp info according to the format.
  * set using qb_log_format_set()
  * It is intended to be used from your custom logger function.
  */
 void qb_log_target_format(int32_t target,
 			  struct qb_log_callsite *cs,
-			  time_t timestamp,
+			  struct timespec *timestamp,
 			  const char* formatted_message,
 			  char *output_buffer);
 
