@@ -71,7 +71,7 @@ set_sock_addr(struct sockaddr_un *address, const char *socket_name)
 	address->sun_len = QB_SUN_LEN(address);
 #endif
 
-	if (!use_filesystem_sockets()) {
+	if (socket_name[0] == '/' || !use_filesystem_sockets()) {
 		snprintf(address->sun_path + 1, UNIX_PATH_MAX - 1, "%s", socket_name);
 	} else {
 		snprintf(address->sun_path, sizeof(address->sun_path), "%s/%s", SOCKETDIR,
@@ -108,8 +108,8 @@ qb_ipc_dgram_sock_setup(const char *base_name,
 		   sizeof(local_address));
 
 	if (use_filesystem_sockets()) {
-		chmod(local_address.sun_path, 0660);
-		chown(local_address.sun_path, -1, gid);
+		(void)chmod(local_address.sun_path, 0660);
+		(void)chown(local_address.sun_path, -1, gid);
 	}
 	if (res < 0) {
 		goto error_connect;
