@@ -23,6 +23,7 @@
 
 #include "os_base.h"
 #include <sys/wait.h>
+#include <sys/un.h>
 #include <signal.h>
 #include <stdbool.h>
 #include <fcntl.h>
@@ -43,6 +44,8 @@
 #include "_failure_injection.h"
 #endif
 
+#define NUM_STRESS_CONNECTIONS 5000
+
 static char ipc_name[256];
 
 #define DEFAULT_MAX_MSG_SIZE (8192*16)
@@ -187,6 +190,8 @@ static GSourceFuncs gio_source_funcs = {
  * 7) service availability
  *
  * 8) multiple services
+ *
+ * 9) setting perms on the sockets
  */
 static qb_loop_t *my_loop;
 static qb_ipcs_service_t* s1;
@@ -198,6 +203,7 @@ static int32_t num_bulk_events = 10;
 static int32_t num_stress_events = 30000;
 static int32_t reference_count_test = QB_FALSE;
 static int32_t multiple_connections = QB_FALSE;
+static int32_t set_perms_on_socket = QB_FALSE;
 
 
 static int32_t
@@ -211,26 +217,29 @@ exit_handler(int32_t rsignal, void *data)
 static void
 set_ipc_name(const char *prefix)
 {
-	/* We have to make the server name as unique as possible given
-	 * the build- (seconds part of preprocessor's timestamp) and
-	 * run-time (pid + lower 16 bits of the current timestamp)
-	 * circumstances, because some build systems attempt to generate
-	 * packages for libqb in parallel.  These unit tests are run
-	 * during the package build process.  2+ builds executing on
-	 * the same machine (whether containerized or not because of
-	 * abstract unix sockets namespace sharing) can stomp on each
-	 * other's unit tests if the ipc server names aren't unique... */
+	FILE *f;
+	char process_name[256];
 
-	/* single-shot grab of seconds part of preprocessor's timestamp */
-	static char t_sec[3] = "";
-	if (t_sec[0] == '\0') {
-		const char *const found = strrchr(__TIME__, ':');
-		strncpy(t_sec, found ? found + 1 : "-", sizeof(t_sec) - 1);
-		t_sec[sizeof(t_sec) - 1] = '\0';
+	/* The process-unique part of the IPC name has already been decided
+	 * and stored in the file ipc-test-name.
+	 */
+	f = fopen("ipc-test-name", "r");
+	if (f) {
+		fgets(process_name, sizeof(process_name), f);
+		fclose(f);
+		snprintf(ipc_name, sizeof(ipc_name), "%.44s%s", prefix, process_name);
+	} else {
+		/* This is the old code, use only as a fallback */
+		static char t_sec[3] = "";
+		if (t_sec[0] == '\0') {
+			const char *const found = strrchr(__TIME__, ':');
+			strncpy(t_sec, found ? found + 1 : "-", sizeof(t_sec) - 1);
+			t_sec[sizeof(t_sec) - 1] = '\0';
+		}
+
+		snprintf(ipc_name, sizeof(ipc_name), "%.44s%s%lX%.4x", prefix, t_sec,
+			 (unsigned long)getpid(), (unsigned) ((long) time(NULL) % (0x10000)));
 	}
-
-	snprintf(ipc_name, sizeof(ipc_name), "%.44s%s%lX%.4x", prefix, t_sec,
-		 (unsigned long)getpid(), (unsigned) ((long) time(NULL) % (0x10000)));
 }
 
 static int
@@ -600,6 +609,13 @@ s1_connection_closed(qb_ipcs_connection_t *c)
 	if (multiple_connections) {
 		return 0;
 	}
+	/* Stop the connection being freed when we call qb_ipcs_disconnect()
+	   in the callback */
+	if (disconnect_after_created == QB_TRUE) {
+		disconnect_after_created = QB_FALSE;
+		return 1;
+	}
+
 	qb_enter();
 	qb_leave();
 	return 0;
@@ -653,6 +669,16 @@ s1_connection_destroyed(qb_ipcs_connection_t *c)
 	qb_leave();
 }
 
+static int32_t
+s1_connection_accept(qb_ipcs_connection_t *c, uid_t uid, gid_t gid)
+{
+	if (set_perms_on_socket) {
+		qb_ipcs_connection_auth_set(c, 555, 741, S_IRWXU|S_IRWXG|S_IROTH|S_IWOTH);
+	}
+	return 0;
+}
+
+
 static void
 s1_connection_created(qb_ipcs_connection_t *c)
 {
@@ -691,9 +717,9 @@ s1_connection_created(qb_ipcs_connection_t *c)
 static volatile sig_atomic_t usr1_bit;
 
 static void usr1_bit_setter(int signal) {
-	if (signal == SIGUSR1) {
-		usr1_bit = 1;
-	}
+    if (signal == SIGUSR1) {
+        usr1_bit = 1;
+    }
 }
 
 #define READY_SIGNALLER(name, data_arg)  void (name)(void *data_arg)
@@ -717,7 +743,7 @@ NEW_PROCESS_RUNNER(run_ipc_server, ready_signaller, signaller_data, data)
 	qb_loop_signal_handle handle;
 
 	struct qb_ipcs_service_handlers sh = {
-		.connection_accept = NULL,
+		.connection_accept = s1_connection_accept,
 		.connection_created = s1_connection_created,
 		.msg_process = s1_msg_process_fn,
 		.connection_destroyed = s1_connection_destroyed,
@@ -790,8 +816,7 @@ run_function_in_new_process(const char *role,
                             void *data)
 {
 	char formatbuf[1024];
-	pid_t parent_target, pid;
-
+	pid_t parent_target, pid1, pid2;
 	struct sigaction orig_sa, purpose_sa;
 	sigset_t orig_mask, purpose_mask, purpose_clear_mask;
 
@@ -806,27 +831,32 @@ run_function_in_new_process(const char *role,
 	purpose_sa.sa_mask = purpose_mask;
 	purpose_sa.sa_flags = SA_RESTART;
 
+	/* Double-fork so the servers can be reaped in a timely manner */
 	parent_target = getpid();
-	pid = fork();
-
-	if (pid == -1) {
-		fprintf (stderr, "Can't fork\n");
-		return -1;
-	}
-
-	if (pid == 0) {
-		sigprocmask(SIG_SETMASK, &orig_mask, NULL);
-
-		if (role == NULL) {
-			qb_log_format_set(QB_LOG_STDERR, "lib/%f|%l[%P] %b");
-		} else {
-			snprintf(formatbuf, sizeof(formatbuf),
-			         "lib/%%f|%%l|%s[%%P] %%b", role);
-			qb_log_format_set(QB_LOG_STDERR, formatbuf);
+	pid1 = fork();
+	if (pid1 == 0) {
+		pid2 = fork();
+		if (pid2 == -1) {
+			fprintf (stderr, "Can't fork twice\n");
+			exit(0);
 		}
+		if (pid2 == 0) {
+			sigprocmask(SIG_SETMASK, &orig_mask, NULL);
 
-		new_process_runner(usr1_signaller, &parent_target, data);
-		exit(0);
+			if (role == NULL) {
+				qb_log_format_set(QB_LOG_STDERR, "lib/%f|%l[%P] %b");
+			} else {
+				snprintf(formatbuf, sizeof(formatbuf),
+				         "lib/%%f|%%l|%s[%%P] %%b", role);
+				qb_log_format_set(QB_LOG_STDERR, formatbuf);
+			}
+
+			new_process_runner(usr1_signaller, &parent_target, data);
+			exit(0);
+		} else {
+			waitpid(pid2, NULL, 0);
+			exit(0);
+		}
 	}
 
 	usr1_bit = 0;
@@ -842,7 +872,7 @@ run_function_in_new_process(const char *role,
 	/* give children a slight/non-strict scheduling advantage */
 	sched_yield();
 
-	return pid;
+	return pid1;
 }
 
 static void
@@ -908,7 +938,7 @@ verify_graceful_stop(pid_t pid)
 	} else {
 		fail_if(rc == 0);
 	}
-	
+
 	return 0;
 }
 
@@ -1406,7 +1436,7 @@ test_ipc_stress_connections(void)
 	pid = run_function_in_new_process("server", run_ipc_server, NULL);
 	fail_if(pid == -1);
 
-	for (connections = 1; connections < 70000; connections++) {
+	for (connections = 1; connections < NUM_STRESS_CONNECTIONS; connections++) {
 		if (conn) {
 			qb_ipcc_disconnect(conn);
 			conn = NULL;
@@ -1421,7 +1451,7 @@ test_ipc_stress_connections(void)
 			}
 		} while (conn == NULL && c < 5);
 		fail_if(conn == NULL);
-		
+
 		if (((connections+1) % 1000) == 0) {
 			qb_log(LOG_INFO, "%d ipc connections made", connections+1);
 		}
@@ -1595,7 +1625,6 @@ END_TEST
 START_TEST(test_ipc_bulk_events_us)
 {
 	qb_enter();
-	send_event_on_created = QB_FALSE;
 	ipc_type = QB_IPC_SOCKET;
 	set_ipc_name(__func__);
 	test_ipc_bulk_events();
@@ -1609,7 +1638,7 @@ READY_SIGNALLER(connected_signaller, _)
 	request_server_exit();
 }
 
-START_TEST(test_ipc_dispatch_us_native_prio_deadlock_provoke)
+START_TEST(test_ipc_dispatch_us_native_prio_dlock)
 {
 	pid_t server_pid, alphaclient_pid;
 	struct dispatch_data data;
@@ -1648,7 +1677,7 @@ START_TEST(test_ipc_dispatch_us_native_prio_deadlock_provoke)
 END_TEST
 
 #if HAVE_GLIB
-START_TEST(test_ipc_dispatch_us_glib_prio_deadlock_provoke)
+START_TEST(test_ipc_dispatch_us_glib_prio_dlock)
 {
 	pid_t server_pid, alphaclient_pid;
 	struct dispatch_data data;
@@ -1881,7 +1910,70 @@ START_TEST(test_ipc_stress_connections_shm)
 }
 END_TEST
 
-START_TEST(test_ipc_dispatch_shm_native_prio_deadlock_provoke)
+// Check perms uses illegal access to libqb internals
+// DO NOT try this at home.
+#include "../lib/ipc_int.h"
+#include "../lib/ringbuffer_int.h"
+START_TEST(test_ipc_server_perms)
+{
+	pid_t pid;
+	struct stat st;
+	int j;
+	uint32_t max_size;
+	int res;
+	int c = 0;
+
+	// Can only test this if we are root
+	if (getuid() != 0) {
+		return;
+	}
+
+	ipc_type = QB_IPC_SHM;
+	set_perms_on_socket = QB_TRUE;
+	max_size = MAX_MSG_SIZE;
+
+	pid = run_function_in_new_process("server", run_ipc_server, NULL);
+	fail_if(pid == -1);
+
+	do {
+		conn = qb_ipcc_connect(ipc_name, max_size);
+		if (conn == NULL) {
+			j = waitpid(pid, NULL, WNOHANG);
+			ck_assert_int_eq(j, 0);
+			poll(NULL, 0, 400);
+			c++;
+		}
+	} while (conn == NULL && c < 5);
+	fail_if(conn == NULL);
+
+	/* Check perms - uses illegal access to libqb internals */
+
+	/* BSD uses /var/run for sockets so we can't alter the perms on the
+	   directory */
+#ifdef __linux__
+	char sockdir[PATH_MAX];
+	strcpy(sockdir, conn->request.u.shm.rb->shared_hdr->hdr_path);
+	*strrchr(sockdir, '/') = 0;
+
+	res = stat(sockdir, &st);
+
+	ck_assert_int_eq(res, 0);
+	ck_assert(st.st_mode & S_IRWXG);
+	ck_assert_int_eq(st.st_uid, 555);
+	ck_assert_int_eq(st.st_gid, 741);
+#endif
+
+	res = stat(conn->request.u.shm.rb->shared_hdr->hdr_path, &st);
+	ck_assert_int_eq(res, 0);
+	ck_assert_int_eq(st.st_uid, 555);
+	ck_assert_int_eq(st.st_gid, 741);
+
+	qb_ipcc_disconnect(conn);
+	verify_graceful_stop(pid);
+}
+END_TEST
+
+START_TEST(test_ipc_dispatch_shm_native_prio_dlock)
 {
 	pid_t server_pid, alphaclient_pid;
 	struct dispatch_data data;
@@ -1920,7 +2012,7 @@ START_TEST(test_ipc_dispatch_shm_native_prio_deadlock_provoke)
 END_TEST
 
 #if HAVE_GLIB
-START_TEST(test_ipc_dispatch_shm_glib_prio_deadlock_provoke)
+START_TEST(test_ipc_dispatch_shm_glib_prio_dlock)
 {
 	pid_t server_pid, alphaclient_pid;
 	struct dispatch_data data;
@@ -1997,11 +2089,15 @@ END_TEST
 START_TEST(test_ipcc_truncate_when_unlink_fails_shm)
 {
 	char sock_file[PATH_MAX];
+	struct sockaddr_un socka;
+
 	qb_enter();
 	ipc_type = QB_IPC_SHM;
 	set_ipc_name(__func__);
 
 	sprintf(sock_file, "%s/%s", SOCKETDIR, ipc_name);
+	sock_file[sizeof(socka.sun_path)] = '\0';
+
 	/* If there's an old socket left from a previous run this test will fail
 	   unexpectedly, so try to remove it first */
 	unlink(sock_file);
@@ -2009,6 +2105,7 @@ START_TEST(test_ipcc_truncate_when_unlink_fails_shm)
 	_fi_unlink_inject_failure = QB_TRUE;
 	test_ipc_server_fail();
 	_fi_unlink_inject_failure = QB_FALSE;
+	unlink(sock_file);
 	qb_leave();
 }
 END_TEST
@@ -2032,7 +2129,7 @@ test_ipc_service_ref_count(void)
 		if (conn == NULL) {
 			j = waitpid(pid, NULL, WNOHANG);
 			ck_assert_int_eq(j, 0);
-			poll(NULL, 0, 400);
+			(void)poll(NULL, 0, 400);
 			c++;
 		}
 	} while (conn == NULL && c < 5);
@@ -2067,7 +2164,7 @@ END_TEST
 #if 0
 static void test_max_dgram_size(void)
 {
-	/* most implementations will not let you set a dgram buffer 
+	/* most implementations will not let you set a dgram buffer
 	 * of 1 million bytes. This test verifies that the we can detect
 	 * the max dgram buffersize regardless, and that the value we detect
 	 * is consistent. */
@@ -2127,10 +2224,11 @@ make_shm_suite(void)
 	add_tcase(s, tc, test_ipc_exit_shm, 6);
 	add_tcase(s, tc, test_ipc_event_on_created_shm, 9);
 	add_tcase(s, tc, test_ipc_service_ref_count_shm, 9);
+	add_tcase(s, tc, test_ipc_server_perms, 7);
 	add_tcase(s, tc, test_ipc_stress_connections_shm, 3600 /* ? */);
-	add_tcase(s, tc, test_ipc_dispatch_shm_native_prio_deadlock_provoke, 15);
+	add_tcase(s, tc, test_ipc_dispatch_shm_native_prio_dlock, 15);
 #if HAVE_GLIB
-	add_tcase(s, tc, test_ipc_dispatch_shm_glib_prio_deadlock_provoke, 15);
+	add_tcase(s, tc, test_ipc_dispatch_shm_glib_prio_dlock, 15);
 #endif
 #ifdef HAVE_FAILURE_INJECTION
 	add_tcase(s, tc, test_ipcc_truncate_when_unlink_fails_shm, 8);
@@ -2163,9 +2261,9 @@ make_soc_suite(void)
 	add_tcase(s, tc, test_ipc_disconnect_after_created_us, 9);
 	add_tcase(s, tc, test_ipc_service_ref_count_us, 9);
 	add_tcase(s, tc, test_ipc_stress_connections_us, 3600 /* ? */);
-	add_tcase(s, tc, test_ipc_dispatch_us_native_prio_deadlock_provoke, 15);
+	add_tcase(s, tc, test_ipc_dispatch_us_native_prio_dlock, 15);
 #if HAVE_GLIB
-	add_tcase(s, tc, test_ipc_dispatch_us_glib_prio_deadlock_provoke, 15);
+	add_tcase(s, tc, test_ipc_dispatch_us_glib_prio_dlock, 15);
 #endif
 
 	return s;
